import shutil
from pathlib import Path

import pytest
from django.conf import settings
from django.contrib.staticfiles.storage import staticfiles_storage
from django.core.cache import caches
from django.core.management import call_command
from django.template.loader import render_to_string

import sri
from sri.algorithm import Algorithm
from sri.templatetags import sri as templatetags

TEST_FILES = ["index.css", "index.js", "admin/js/core.js"]


def setup_function(*_):
    for cache in caches.all():
        cache.clear()  # Clear cache between each test method
    shutil.rmtree(settings.STATIC_ROOT, ignore_errors=True)


def test_simple_template():
    rendered = render_to_string("simple.html")
    assert (
        '<script crossorigin="anonymous" integrity="sha256-VROI/fAMCWgkTthVtzzvHtPkkxvpysdZbcqLdVMtwOI=" src="/static/index.js"></script>'
        in rendered
    )
    assert (
        '<link crossorigin="anonymous" href="/static/index.css" integrity="sha256-fsqAKvNYgo9VQgSc4rD93SiW/AjKFwLtWlPi6qviBxY=" rel="stylesheet" type="text/css">'
        in rendered
    )


def test_complex_template():
    rendered = render_to_string("complex.html")
    assert (
        '<script crossorigin="anonymous" integrity="sha256-VROI/fAMCWgkTthVtzzvHtPkkxvpysdZbcqLdVMtwOI=" src="/static/index.js" defer async></script>'
        in rendered
    )
    assert (
        '<link as="font" crossorigin="anonymous" href="/static/index.woff2" integrity="sha256-hWU2c2zzSsvKYN7tGMnt3t3Oj7GwQZB2aLRhCWYbFSE=" preload>'
        in rendered
    ), rendered


def test_algorithms_template():
    rendered = render_to_string("algorithms.html")
    assert (
        '<script crossorigin="anonymous" integrity="sha384-dExnf54EbXTQ1VmweBEJRWX3MPT4xeDV5p71GIX2hpvV+8B/kzo3SObynuveYt9w" src="/static/index.js"></script>'
        in rendered
    )
    assert (
        '<link crossorigin="anonymous" href="/static/index.css" integrity="sha512-7v9G7AKwpjnlEYhw9GdXu/9G8bq0PqM427/QmgH2TufqEUcjsANEoyCoOkpV8TBCnbQigwNKpMaZNskJG8Ejdw==" rel="stylesheet" type="text/css">'
        in rendered
    )


def test_jinja2_template():
    rendered = render_to_string("complex.j2")
    assert (
        '<script crossorigin="anonymous" integrity="sha256-VROI/fAMCWgkTthVtzzvHtPkkxvpysdZbcqLdVMtwOI=" src="/static/index.js" defer async></script>'
        in rendered
    )
    assert (
        '<link as="font" crossorigin="anonymous" href="/static/index.woff2" integrity="sha256-hWU2c2zzSsvKYN7tGMnt3t3Oj7GwQZB2aLRhCWYbFSE=" preload>'
        in rendered
    )


@pytest.mark.parametrize("algorithm", sri.Algorithm)
@pytest.mark.parametrize("file", TEST_FILES)
def test_generic_algorithm(algorithm, file):
    val = templatetags.sri_integrity_static(file, algorithm)
    assert val.startswith(f"{algorithm.value}-"), val


def test_default_algorithm():
    val = templatetags.sri_integrity_static("index.js")
    assert val.startswith(f"{Algorithm.get_default().value}-"), val


@pytest.mark.parametrize("file", TEST_FILES)
def test_get_static_path(file):
    file_path = sri.utils.get_static_path(file)

    assert file_path.exists()
    assert file_path.is_file()

    if "site-packages" not in str(file_path):
        assert file_path == Path("tests/static").joinpath(file).resolve()


def test_default_algorithm_exists():
    assert Algorithm.get_default() in sri.hashers.HASHERS


@pytest.mark.parametrize("algorithm", sri.Algorithm)
@pytest.mark.parametrize("file", TEST_FILES)
def test_hashes_are_consistent(algorithm, file):
    digest = sri.hashers.calculate_hash(sri.utils.get_static_path(file), algorithm)
    caches["default"].clear()
    digest_2 = sri.hashers.calculate_hash(sri.utils.get_static_path(file), algorithm)
    assert digest == digest_2


@pytest.mark.parametrize("algorithm", sri.Algorithm)
@pytest.mark.parametrize("file", TEST_FILES)
def test_integrity(algorithm, file):
    integrity = sri.integrity.calculate_integrity(
        sri.utils.get_static_path(file), algorithm
    )
    assert integrity.startswith(algorithm.value)


@pytest.mark.parametrize("file", TEST_FILES)
def test_disable_sri(file):
    original_value = templatetags.USE_SRI
    try:
        templatetags.USE_SRI = False
        assert "integrity" not in templatetags.sri_static(file)
    finally:
        templatetags.USE_SRI = original_value


@pytest.mark.parametrize("algorithm", sri.Algorithm)
@pytest.mark.parametrize("file", TEST_FILES)
def test_sri_integrity_static(algorithm, file):
    assert templatetags.sri_integrity_static(file, algorithm).startswith(
        f"{algorithm.value}-"
    )


@pytest.mark.parametrize("file", TEST_FILES)
def test_unknown_algorithm(file):
    with pytest.raises(ValueError) as e:
        templatetags.sri_static(file, algorithm="md5")
    assert e.value.args[0] == "'md5' is not a valid Algorithm"


def test_missing_file():
    with pytest.raises(FileNotFoundError):
        templatetags.sri_static("foo.js")


def test_app_file():
    templatetags.sri_static("admin/js/core.js")


@pytest.mark.parametrize("algorithm", sri.Algorithm)
@pytest.mark.parametrize("file", TEST_FILES)
def test_caches_hash(algorithm, file):
    file_path = sri.utils.get_static_path(file)
    cache_key = sri.hashers.get_cache_key(file_path, algorithm)
    cache = caches["default"]

    assert cache.get(cache_key) is None
    digest = sri.hashers.calculate_hash(file_path, algorithm)
    assert cache.get(cache_key) == digest


@pytest.mark.parametrize("file", TEST_FILES)
def test_manifest_storage(settings, file):
    settings.STORAGES = {
        "staticfiles": {
            "BACKEND": "django.contrib.staticfiles.storage.ManifestStaticFilesStorage"
        }
    }
    call_command("collectstatic", interactive=False, clear=True, verbosity=0)

    file_path = sri.utils.get_static_path(file)

    assert file_path.exists()
    assert file_path.is_file()

    assert str(file_path).startswith(settings.STATIC_ROOT)
    assert not str(file_path).endswith(file)
    assert str(file_path).endswith(staticfiles_storage.stored_name(file))


@pytest.mark.parametrize("file", TEST_FILES)
def test_default_storage(file):
    # Test for issue #70
    call_command("collectstatic", interactive=False, clear=True, verbosity=0)

    file_path = sri.utils.get_static_path(file)

    assert file_path.exists()
    assert file_path.is_file()
    # If you rollback the changes outlined in issue #70, this
    # will fail as the path returned will be the source path
    # and not the destination path.
    assert str(file_path).startswith(settings.STATIC_ROOT)


@pytest.mark.parametrize(
    "empty,extra,output",
    [
        ([], {}, ""),
        (["defer"], {}, " defer"),
        (["defer", "async"], {}, " defer async"),
        ([], {"type": "text/javascript"}, ' type="text/javascript"'),
        (["defer"], {"type": "text/javascript"}, ' type="text/javascript" defer'),
    ],
)
def test_format_attrs(empty, extra, output: str) -> None:
    elem = templatetags.format_attrs(*empty, **extra)
    assert elem == output, elem



import os.path
from typing import Optional

from django import template
from django.conf import settings
from django.forms.utils import flatatt
from django.templatetags.static import static
from django.utils.safestring import mark_safe

from sri.algorithm import Algorithm
from sri.integrity import calculate_integrity_of_static

USE_SRI = getattr(settings, "USE_SRI", not settings.DEBUG)

register = template.Library()


def format_attrs(*empty_tag_attrs, **extra_tag_attrs) -> str:
    """
    Flatten and format list and dict params.

    The returned value will be a string with a leading " " that is a
    combination of "name=value" and "name" params:

        ' src="/" defer'
        ' href="/" rel="stylesheet"'

    """
    extra_attrs = flatatt(extra_tag_attrs)
    if empty_tag_attrs:
        empty_attrs = " ".join(empty_tag_attrs)
        return f"{extra_attrs} {empty_attrs}".rstrip()
    return extra_attrs.rstrip()


def sri_js(path: str, *empty_tag_attrs: str, **extra_tag_attrs: str) -> str:
    return script_tag(path, *empty_tag_attrs, **extra_tag_attrs)


def sri_css(path: str, *empty_tag_attrs: str, **extra_tag_attrs: str) -> str:
    extra_tag_attrs.setdefault("rel", "stylesheet")
    extra_tag_attrs.setdefault("type", "text/css")
    return link_tag(path, *empty_tag_attrs, **extra_tag_attrs)


def script_tag(path: str, *empty_tag_attrs: str, **extra_tag_attrs: str) -> str:
    extra_tag_attrs.setdefault("src", static(path))
    return mark_safe(
        f"<script{format_attrs(*empty_tag_attrs, **extra_tag_attrs)}></script>"
    )


def link_tag(path: str, *empty_tag_attrs: str, **extra_tag_attrs: str) -> str:
    extra_tag_attrs.setdefault("href", static(path))
    return mark_safe(f"<link{format_attrs(*empty_tag_attrs, **extra_tag_attrs)}>")


EXTENSIONS = {"js": sri_js, "css": sri_css}


@register.simple_tag
def sri_static(
    path: str,
    *empty_tag_attrs: str,
    algorithm: Optional[str] = None,
    **extra_tag_attrs: str,
) -> str:
    extension = os.path.splitext(path)[1][1:]
    sri_method = EXTENSIONS.get(extension, link_tag)

    algorithm_type = (
        Algorithm(algorithm) if algorithm is not None else Algorithm.get_default()
    )

    if USE_SRI:
        extra_tag_attrs.setdefault("crossorigin", "anonymous")
        extra_tag_attrs["integrity"] = sri_integrity_static(path, algorithm_type)
    return sri_method(path, *empty_tag_attrs, **extra_tag_attrs)


@register.simple_tag
def sri_integrity_static(path: str, algorithm: Optional[str] = None) -> str:
    return calculate_integrity_of_static(
        path, Algorithm(algorithm) if algorithm is not None else Algorithm.get_default()
    )



import logging
import os
from pathlib import Path

from django.contrib.staticfiles.finders import find as find_static_file
from django.contrib.staticfiles.storage import staticfiles_storage

logger = logging.getLogger(__name__)


def get_static_path(path: str) -> Path:
    """
    Resolves a path commonly passed to `{% static %}` into a filesystem path
    """

    if hasattr(staticfiles_storage, "stored_name"):
        path = staticfiles_storage.stored_name(path)

    collected_file_path = staticfiles_storage.path(path)
    if os.path.exists(collected_file_path):
        return Path(collected_file_path)

    logger.debug("File not found in staticfiles_storage - checking source files")
    source_static_file_path = find_static_file(path)
    if source_static_file_path is not None:
        return Path(source_static_file_path)

    raise FileNotFoundError(path)



import base64
import hashlib
from functools import lru_cache
from pathlib import Path

from django.core.cache import DEFAULT_CACHE_ALIAS, caches
from django.core.cache.backends.base import InvalidCacheBackendError

from sri.algorithm import Algorithm

HASHERS = {
    Algorithm.SHA256: hashlib.sha256,
    Algorithm.SHA384: hashlib.sha384,
    Algorithm.SHA512: hashlib.sha512,
}

READ_BUFFER_SIZE = 2**18  # 256k - matches `hashlib.file_digest`


def calculate_hash(path: Path, algorithm: Algorithm) -> str:
    try:
        cache = caches["sri"]
    except InvalidCacheBackendError:
        cache = caches[DEFAULT_CACHE_ALIAS]

    cache_key = get_cache_key(path, algorithm)
    file_hash = cache.get(cache_key)
    if file_hash is None:
        # Cache miss, do the calculation
        with path.open("rb") as f:
            if hasattr(hashlib, "file_digest"):
                hasher = hashlib.file_digest(f, HASHERS[algorithm])
            else:
                hasher = HASHERS[algorithm]()
                while True:
                    data = f.read(READ_BUFFER_SIZE)
                    if not data:
                        break
                    hasher.update(data)
        file_hash = base64.b64encode(hasher.digest()).decode()
        cache.set(cache_key, file_hash)
    return file_hash


@lru_cache(maxsize=None)
def get_cache_key(path: Path, algorithm: Algorithm) -> str:
    path_hash = hashlib.sha1(str(path).encode(), usedforsecurity=False).hexdigest()
    return f"sri-{path_hash}-{algorithm.value}"



from jinja2 import Environment
from jinja2.ext import Extension

from sri.templatetags.sri import sri_integrity_static, sri_static


class SRIExtension(Extension):
    def __init__(self, environment: Environment) -> None:
        super().__init__(environment)

        environment.globals["sri_static"] = sri_static
        environment.globals["sri_integrity_static"] = sri_integrity_static


# Shorthand
sri = SRIExtension



from enum import Enum, unique

from django.conf import settings


@unique
class Algorithm(Enum):
    SHA256 = "sha256"
    SHA384 = "sha384"
    SHA512 = "sha512"

    @classmethod
    def get_default(cls) -> "Algorithm":
        return Algorithm(getattr(settings, "SRI_ALGORITHM", Algorithm.SHA256))



from setuptools import find_packages, setup

with open("README.md") as f:
    readme = f.read()

setup(
    name="django-sri",
    version="0.8.0",
    url="https://github.com/RealOrangeOne/django-sri",
    author="Jake Howard",
    description="Subresource Integrity for Django",
    long_description=readme,
    long_description_content_type="text/markdown",
    license="BSD",
    packages=find_packages(include=["sri*"]),
    package_data={"sri": ["py.typed"]},
    install_requires=["Django>=4.2"],
    python_requires=">=3.9",
    keywords="django subresource integrity sri",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 4.2",
        "Framework :: Django :: 5.0",
        "Framework :: Django :: 5.1",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development",
        "Typing :: Typed",
    ],
)



from pathlib import Path
from typing import Optional

from sri.algorithm import Algorithm
from sri.hashers import calculate_hash
from sri.utils import get_static_path


def calculate_integrity(path: Path, algorithm: Optional[Algorithm] = None) -> str:
    if algorithm is None:
        algorithm = Algorithm.get_default()

    return f"{algorithm.value}-{calculate_hash(path, algorithm)}"


def calculate_integrity_of_static(
    static_path: str, algorithm: Optional[Algorithm] = None
) -> str:
    return calculate_integrity(get_static_path(static_path), algorithm)
